import path from 'path';
import now from 'performance-now';
import * as env from 'dotenv';
env.config();
const Knex = require('knex');

export const connection = {
  host: process.env.DB_MASTER_HOST,
  user: process.env.DB_MASTER_USER,
  password: process.env.DB_MASTER_PASS,
  database: process.env.DB_MASTER_NAME,
};

export const knexConfigObj = Knex({
  client: 'mysql2',
  useNullAsDefault: true,
  connection: connection,
  debug: true,
  pool: { min: 0, max: 50 },
  migrations: {
    directory: path.join(__dirname, '../database/database.ts'),
  },
  seeds: {
    directory: path.join(__dirname, '../database/seeds/'),
  },
});

export const knexMaster = Knex({
  client: 'mysql2',
  useNullAsDefault: true,
  connection: connection,
  debug: true,
  pool: { min: 0, max: 50 },
});

export const knexRead = Knex({
  client: 'mysql2',
  useNullAsDefault: true,
  connection: connection,
  debug: true,
  pool: { min: 0, max: 50 },
});
const times: any = {};
// Used for keeping track of the order queries are executed.
let count = 0;
if (process.env.LOG_CONSOLE_QUERY == 'true') {
  knexMaster
    .on('query', (query: any) => {
      const uid = query.__knexQueryUid;
      times[uid] = {
        position: count,
        query,
        startTime: now(),
        finished: false,
      };
      count = count + 1;
    })
    .on('query-response', (response: any, query: any) => {
      const uid = query.__knexQueryUid;
      times[uid].endTime = now();
      times[uid].finished = true;
      const position = times[uid].position;

      // Print the current query, if I'm able
      printIfPossible(uid);
      printQueriesAfterGivenPosition(position);
    });
  knexRead
    .on('query', (query: any) => {
      const uid = query.__knexQueryUid;
      times[uid] = {
        position: count,
        query,
        startTime: now(),
        finished: false,
      };
      count = count + 1;
    })
    .on('query-response', (response: any, query: any) => {
      const uid = query.__knexQueryUid;
      times[uid].endTime = now();
      times[uid].finished = true;
      const position = times[uid].position;

      // Print the current query, if I'm able
      printIfPossible(uid);
      // Check to see if queries further down the queue can be executed,
      //in case they weren't able to be printed when they first responded.
      printQueriesAfterGivenPosition(position);
    });
}

export const knexWrapper = Knex({
  client: 'mysql2',
});

knexWrapper.client.runner = function (builder: any) {
  if (builder._queryContext && builder._queryContext.useMaster === true) {
    return knexMaster.client.runner(builder);
  } else if (builder._queryContext && builder._queryContext.useMaster === false) {
    return knexRead.client.runner(builder);
  }
  const sql = builder.toSQL();
  if (process.env.LOG_CONSOLE_QUERY == 'true') {
    console.log(sql);
  }

  // (knex.select('*').queryContext({ useMaster: true }))
  const useMaster = Array.isArray(sql) ? checkUseMasterMultiple(sql) : checkUseMasterSingle(sql);
  return useMaster ? knexMaster.client.runner(builder) : knexRead.client.runner(builder);
};

const checkUseMasterSingle = (object: any) => {
  return ['insert', 'del', 'update'].includes(object.method);
};

// this is just a simple loop
const checkUseMasterMultiple = (array: any) => {
  for (let i = 0; i < array.length; i++) {
    if (checkUseMasterSingle(array[i])) {
      return true;
    }
  }
  return false;
};

const printQueryWithTime = (uid: any) => {
  const { startTime, endTime, query } = times[uid];
  const elapsedTime = endTime - startTime;

  // I print the sql generated for a given query, as well as
  // the bindings for the queries.
  console.log(query.sql, ',', `[${query.bindings ? query.bindings.join(',') : ''}]`);
  console.log(`Time: ${elapsedTime.toFixed(3)} ms\n`);

  // After I print out the query, I have no more use to it,
  // so I delete it from my map so it doesn't grow out of control.
  delete times[uid];
};
const printIfPossible = (uid: any) => {
  const { position } = times[uid];

  // Look of a query with a position one less than the current query
  const previousTimeUid = Object.keys(times).find(key => times[key].position === position - 1);

  // If we didn't find it, it must have been printed already and we can safely print ourselves.
  if (!previousTimeUid) {
    printQueryWithTime(uid);
  }
};

const printQueriesAfterGivenPosition = (position: any) => {
  // Look for the next query in the queue
  const nextTimeUid = Object.keys(times).find(key => times[key].position === position + 1);

  // If we find one and it is marked as finished, we can go ahead and print it
  if (nextTimeUid && times[nextTimeUid].finished) {
    const nextPosition = times[nextTimeUid].position;
    printQueryWithTime(nextTimeUid);

    // There might be more queries that need to printed, so we should keep looking...
    printQueriesAfterGivenPosition(nextPosition);
  }
};
